import sys,time,functools
from ..viewer import ImageDrawer#,ImageObjectFactory,ArchiveViewer

try:
    import pygtk
    pygtk.require('2.0')
    import gtk,gobject
    import gtk.glade
except:
    print sys.exc_info()
    print 'Import Error:pygtk,gtk,gtk.glade'
    sys.exit(1)

class ComicBrowser:
    __is_sleeping = False
    __cur_size = 0,0
    __box_margin = None
    viewer = None
    drawer = None
    __prev_pressed_key = None
    __omitted_func = None
    __key_repeat = 0
    __slideshow_thread = None
    __forced_slideshow_interval = 250
    __zoom_rate_degree = 0.1
    __is_slideshow_running = False
    __is_forced_slideshow_running = False

    def __init__(self):
        self.gladefile = 'viewer.glade'
        self.wTree = gtk.glade.XML(self.gladefile)
        self.wTree.signal_autoconnect(self)
        self.wmain = self.wTree.get_widget("MainWindow")
        self.imgbox = self.wTree.get_widget("ImageContainer")
        self.pagejumpdialog = self.wTree.get_widget('PageJumpDialog')
        self.chooser = self.wTree.get_widget('ArchiveFileChooserDialog')
        self.evtbox_left = self.wTree.get_widget("PageImageEventBoxLeft")
        self.menubar = self.wTree.get_widget('MainMenuBar')
        self.imgbox_adjusts = self.imgbox.get_hadjustment(),self.imgbox.get_vadjustment()
        self.pagespin = self.wTree.get_widget('PageNumSpinButton')
        self.drawer = ImageDrawer(self.wTree.get_widget("PageImageRight"),
                                  self.wTree.get_widget("PageImageLeft"))
        self.img_factory = ImageObjectFactory(GImageObject)
        self.__set_bg_color()
        
    def slice_time(func):
        #switching process time for threads
        #sometimes,threads are locked or stoped by GIL
        @functools.wraps(func)
        def wrapper(*args,**kw):
            time.sleep(0.001)
            return func(*args,**kw)
        return wrapper

    def __wake_up_from_slideshow(self):
        if self.__is_slideshow_running:
            self.__is_slideshow_running = False

    def __set_bg_color(self):
        color = gtk.gdk.color_parse(BG_COLOR)
        self.imgbox.modify_bg(gtk.STATE_NORMAL,
                              color)

    def __install_chooser_filter(self,extensions):
        def create_filter(pattern,name):
            ffilter = gtk.FileFilter()
            ffilter.set_name(name)
            ffilter.add_pattern(pattern)
            return ffilter

        chooser = self.chooser
        for old_filter in chooser.list_filters():
            chooser.remove_filter(old_filter)
        
        for name in extensions.keys():
            pattern = '|'.join(['*'+ extn for extn in  extensions[name]])
            ffilter = create_filter( pattern ,name)
            chooser.add_filter(ffilter)

    def __init_chooser(self,extensions):
        self.__install_chooser_filter(extensions)
        self.chooser.unselect_all()
        self.chooser.resize( gtk.gdk.screen_width() / 2,gtk.gdk.screen_height() /2   ) 

    def __start_slideshow(self,interval,move_func):
        def __activate_slideshow():
            move_func()
            return self.__is_slideshow_running
        self.__is_slideshow_running = True
        gobject.timeout_add(interval, __activate_slideshow)
    
    def __activate_miFitManually(self):
        mi = self.wTree.get_widget('miFitManually')
        if not mi.get_active():
            mi.set_active(True)
        if self.drawer.get_fit_mode() != self.drawer.FIT_MANUALLY:
            self.drawer.set_fit_mode(self.drawer.FIT_MANUALLY)
        else:
            self.drawer.redraw()

    @slice_time
    def on_miSlideshow_activate(self,widget):
        self.__start_slideshow(SLIDESHOW_INTERVAL,self.viewer.move_next_page)

    def on_PageJumpDialog_response(self,widget,response):
        if response == gtk.RESPONSE_OK:
            self.viewer.move_page_to( self.pagespin.get_value() )
        widget.hide()

    @slice_time
    def on_miMoveToInputedPage_activate(self,widget):
        if not self.viewer:return
        spin = self.pagespin ; viewer = self.viewer
        minimum ,maximum = spin.get_range()
        page_sum = viewer.get_page_sum() - 1
        if maximum != page_sum:
            self.wTree.get_widget('PageSumLabel').set_text(str(page_sum))
            spin.set_range(0,page_sum)
            spin.set_value(viewer.get_page_number())

        self.pagejumpdialog.show()

    @slice_time
    def on_miMoveToNext_activate(self,widget):
        self.viewer.move_next_page()

    @slice_time
    def on_miMoveToPrev_activate(self,widget):
        self.viewer.move_prev_page()

    @slice_time
    def on_miMoveToFirst_activate(self,widget):
        self.viewer.move_first_page()

    @slice_time
    def on_miMoveToLast_activate(self,widget):
        self.viewer.move_last_page()

    @slice_time
    def on_miFullscreen_toggled(self,widget):
        if widget.get_active():
            self.wmain.fullscreen()
        else:
            self.wmain.unfullscreen()

    @slice_time
    def on_miRightToLeft_toggled(self,widget):
        self.drawer.set_right_to_left_mode( widget.get_active() )

    @slice_time
    def on_miDualview_toggled(self,widget):
        is_active = widget.get_active()
        if is_active:
            self.evtbox_left.show()
        else:
            self.evtbox_left.hide()
        self.drawer.set_dualscreen_mode(is_active)

    @slice_time
    def on_miFitAuto_toggled(self,widget):
        if widget.get_active():
            self.drawer.set_fit_mode(self.drawer.FIT_AUTO)

    @slice_time
    def on_miFitToWidth_toggled(self,widget):
        if widget.get_active():
            self.drawer.set_fit_mode(self.drawer.FIT_TO_WIDTH)

    @slice_time
    def on_miFitToHeight_toggled(self,widget):
        if widget.get_active():
            self.drawer.set_fit_mode(self.drawer.FIT_TO_HEIGHT)

    @slice_time
    def on_miZoomOrign_activate(self,widget):
        self.drawer.set_manual_zoom_rate(1.0)
        self.__activate_miFitManually()

    @slice_time
    def on_miZoomIn_activate(self,widget):
        current_rate = self.drawer.get_manual_zoom_rate()
        new_rate = current_rate + self.__zoom_rate_degree
        self.drawer.set_manual_zoom_rate( new_rate)
        self.__activate_miFitManually()
        
    @slice_time
    def on_miZoomOut_activate(self,widget):
        current_rate = self.drawer.get_manual_zoom_rate()
        new_rate = current_rate - self.__zoom_rate_degree
        if new_rate < 0.0: return
        self.drawer.set_manual_zoom_rate( new_rate )
        self.__activate_miFitManually()

    def on_miArchiveFileOpen_activate(self,widget):
        self.__init_chooser(archive.get_support_archive())
        self.chooser.show()

    def on_miImageFileOpen_activate(self,widget):
        self.__init_chooser(format.get_support_image())
        self.chooser.show()

    def on_ArchiveFileChooserDialog_response(self,widget,response):
        if response == gtk.RESPONSE_OK:
            file_uri = widget.get_uri()
            self.viewer = ArchiveViewer(self.drawer,self.img_factory)
            self.viewer.run(file_uri)
        widget.hide()

    def on_miAppInfo_activate(self,widget):
        self.wTree.get_widget('AboutAppDialog').show()

    def on_AboutAppDialog_response(self,widget,response):
        widget.hide()

    @slice_time
    def on_PageImageEventBoxLeft_button_press_event(self,widget,event):
        if self.viewer:
            if self.drawer.get_right_to_left_mode():
                self.viewer.move_next_page()
            else:
                self.viewer.move_prev_page()

    @slice_time
    def on_PageImageEventBoxRight_button_press_event(self,widget,event):
        if self.viewer:
            if not self.drawer.get_dualscreen_mode():
                self.viewer.move_next_page()
            else:
                if self.drawer.get_right_to_left_mode():
                    self.viewer.move_next_page()
                else:
                    self.viewer.move_prev_page()

    @slice_time
    def on_MainWindow_size_allocate(self,widget,allocation):
        wm_size = self.wmain.get_size()
        if self.__cur_size !=  wm_size:
            self.__cur_size = wm_size
            hadjust,vadjust = self.imgbox_adjusts
            if self.__box_margin == None:
                self.__box_margin = (wm_size[0] - self.imgbox.allocation.width,
                                     self.menubar.allocation.height)

            box_w,box_h = [wm_size[0] - self.__box_margin[0],
                           wm_size[1] - self.__box_margin[1]]
            self.drawer.set_container_size(box_w,
                                           box_h)
            if self.viewer:
                self.drawer.redraw()

    @slice_time
    def on_MainWindow_window_state_event(self,widget,event):
        max_state = gtk.gdk.WINDOW_STATE_MAXIMIZED
        is_max_before = (event.changed_mask & max_state) != 0
        is_max_now =  (event.new_window_state & max_state) != 0
        if is_max_before or is_max_now:
            self.on_MainWindow_size_allocate(widget,None)

    def on_MainWindow_button_press_event(self,widget,event):
        self.__wake_up_from_slideshow()
        return False

    def on_MainWindow_key_release_event(self,widget,event):
        if self.viewer:
            if self.__is_forced_slideshow_running:
                self.__is_forced_slideshow_running = False
                self.__wake_up_from_slideshow()
            self.__prev_pressed_key = None

    def on_MainWindow_key_press_event(self,widget,event):
        is_key_repeat = False

        #if key-press-event repeat until emitting release-event,
        #the move-func works as slideshow,and omit the func hereafter
        def __delegate_move_func(func,*args):
            if is_key_repeat:
                self.__is_forced_slideshow_running = True
                self.__start_slideshow(self.__forced_slideshow_interval,
                                       functools.partial(func,args))
            else:
                func(*args)

        if self.viewer:
            if self.__is_forced_slideshow_running:
                return

            self.__wake_up_from_slideshow()
            
            #compare with previous key and currrent key for key-repeat-check
            key_info = {event.keyval:event.state}
            if self.__prev_pressed_key != key_info:
                self.__prev_pressed_key = key_info
            else:
                is_key_repeat = True

            #here from compare key conditions
            keyval =  event.keyval
            is_ctrl = not (event.state & gtk.gdk.CONTROL_MASK) == 0
            hadjust,vadjust = self.imgbox_adjusts
            if keyval == gtk.keysyms.Right:
                if is_ctrl:
                    self.__on_Scrollbar_changed(hadjust,
                                                True)
                else:
                    __delegate_move_func(self.on_miMoveToNext_activate,(widget,))
            elif keyval == gtk.keysyms.Left:
                if is_ctrl:
                    self.__on_Scrollbar_changed(hadjust,
                                                False)
                else:
                    __delegate_move_func(self.on_miMoveToPrev_activate,(widget,))
            elif keyval == gtk.keysyms.Up:
                if is_ctrl:
                    self.__on_Scrollbar_changed(vadjust,
                                                False)
            elif keyval == gtk.keysyms.Down:
                if is_ctrl:
                    self.__on_Scrollbar_changed(vadjust,
                                                True)

    @slice_time
    def __on_Scrollbar_changed(self,adjust,is_increased):
        if not adjust.upper == 0:
            increment = adjust.step_increment
            newval = 0
            if is_increased:
                newval = adjust.value + increment
                max = adjust.upper - adjust.page_size
                if newval > max:
                    newval = max
            else:
                newval = adjust.value - increment
                if newval < 0:
                    newval = 0
            adjust.value = newval

    def on_miAppQuit_activate(self,widget):
        self.on_MainWindow_destroy(widget)

    def on_MainWindow_destroy(self,widget):
        if self.viewer:
            self.viewer.terminate()
        gtk.main_quit()

    def main(self):
        width = gtk.gdk.screen_width() /2
        height = gtk.gdk.screen_height() /2
        self.wmain.resize(width,height)
        gtk.main()

class GImageObject(object):
    __size = None
    __build_wait_time = 0.0005
    def __init__(self,data):
        self.pixbuf = None
        loader = gtk.gdk.PixbufLoader()
        thread.start_new_thread(self.__write_data,(loader,data,))

    def __write_data(self,loader,data):
        if not loader.write(data):
            raise 'io error or someting...'
        self.pixbuf = loader.get_pixbuf()
        while not self.pixbuf:
            self.pixbuf = loader.get_pixbuf()
            time.sleep(self.__build_wait_time)
        loader.close()
        del loader

    def __sync_building_pixbuf(self):
        while not self.pixbuf:
            time.sleep(self.__build_wait_time)

    @property
    def size(self):
        if self.__size:
            return self.__size
        self.__sync_building_pixbuf()
        self.__size = self.pixbuf.get_width(),self.pixbuf.get_height()
        return self.__size

    def resize(self,size,mode=gtk.gdk.INTERP_BILINEAR):
        self.__sync_building_pixbuf()
        width,height = size
        return self.pixbuf.scale_simple(width,height,mode)
