import threading,time,collections
from . import archive
from . import config
class ImageConverter:
    def to_image_object(self,buffer):
        p = ImageFile.Parser()
        while True:
            dat = buffer.read(1024)
            if not dat:
                break
            p.feed(dat)
        img = p.close()
        return img
    def get_aspect_ratio(self,size):
        def get_common_factor(a,b):
            if b == 0:
                return a
            return get_common_factor(b,a % b)
        width,height = size
        result = get_common_factor(width,height)
        return [width / result,
                height /result]
    def resize(self,img_obj,new_size):
        return img_obj.resize(new_size,
                          Image.ANTIALIAS)
    def create_thumbnail(self,img,size):
        img_thumbnail = img.copy()
        img_thumbnail.thumbnail(size,Image.ANTIALIAS)
        return img_thumbnail
    def to_gtk_picbuf(self,img_obj):
        buf = cStringIO.StringIO()
        img_obj.save(buf,'ppm')
        img_pixs = buf.getvalue()
        buf.close()
        loader = gtk.gdk.PixbufLoader('pnm')
        loader.write(img_pixs,len(img_pixs))
        pixbuf = loader.get_pixbuf()
        loader.close()
        return pixbuf

class ImageObjectFactory:
    def __init__(self,img_class_type):
        self.img_class_type = img_class_type
    def build(self,data):
        return self.img_class_type(data)

class ImageDrawer:
    FIT_AUTO = 0
    FIT_MANUALLY = 1
    FIT_TO_WIDTH = 2
    FIT_TO_HEIGHT = 3
    __cur_img = None
    __fwd_img = None
    __box_size =0,0
    __container_size = 0,0
    __zoom_rate = 0.0

    def __init__(self,img_frameR,img_frameL):
        self.conv = ImageConverter()
        self.img_frameR = img_frameR
        self.img_frameL = img_frameL
        self.__is_dual = False
        self.__fit_mode = self.FIT_AUTO
        self.__is_right_to_left = True
        self.set_dualscreen_mode(self.__is_dual)

    def set_container_size(self,width,height):
        self.__container_size = width,height
        self.__update_box_size()

    def __update_box_size(self):
        self.__box_size = self.__container_size
        if self.__is_dual:
            self.__box_size = self.__box_size[0] / 2,self.__box_size[1]

    def __update_box_alignment(self):
        minimum=0.0 ; center=0.5 ; maximum=1.0 
        frameL_align = center,center
        frameR_align = center,center
        if self.__is_dual:
            if not self.__is_right_to_left:
                frameL_align = minimum,center
                frameR_align = maximum,center
            else:
                frameL_align = maximum,center
                frameR_align = minimum,center
        for frame,align in [[self.img_frameL,frameL_align],
                            [self.img_frameR,frameR_align]]:
            frame.set_alignment(align[0],align[1])

    def redraw(self):
        cur = self.__cur_img ; fwd = self.__fwd_img
        if self.__is_dual:
            img_set = [self.img_frameL,fwd],[self.img_frameR,cur]
            if not self.__is_right_to_left:
                img_set = [self.img_frameR,fwd],[self.img_frameL,cur]

            for frame,img in img_set:
                self.__draw_frame(frame,
                                  self.__resize_img(img))
        else:
            self.__draw_frame(self.img_frameR, self.__resize_img(cur))

    def draw_image(self,img,fwd=None):
        self.__cur_img = img ; self.__fwd_img = fwd
        self.redraw()

    def __draw_frame(self,frame,img):
         frame.set_from_pixbuf(img)

    def __resize_img(self,img,zoom_rate=None):
        resized_size = None
        box_size = self.__box_size ; fit_mode = self.__fit_mode
        ratio = self.conv.get_aspect_ratio(img.size)

        def get_height_mode_size():
            con_h = box_size[1]
            factor = complex( con_h) / ratio[1]
            resized_width = int(abs(ratio[0] * factor))
            zoom_rate = abs(complex(resized_width) / img.size[0])
            return zoom_rate,[resized_width,con_h]
        
        def get_width_mode_size():
            con_w = box_size[0]
            factor = complex(con_w) / ratio[0]
            resized_height = int(abs(ratio[1] * factor))
            zoom_rate = abs(complex(resized_height) / img.size[1])
            return zoom_rate,[con_w,resized_height]
        if fit_mode == self.FIT_AUTO:
            con_w,con_h = box_size
            hzoom_rate,hmode = get_height_mode_size()
            wzoom_rate,wmode = get_width_mode_size()
            if hmode < wmode:
                resized_size = hmode
                self.__zoom_rate = hzoom_rate
            else:
                resized_size = wmode
                self.__zoom_rate = wzoom_rate
        elif fit_mode == self.FIT_TO_HEIGHT:
            zoom_rate,resized_size = get_height_mode_size()
            self.__zoom_rate = zoom_rate
        elif fit_mode == self.FIT_TO_WIDTH:
            zoom,resized_size = get_width_mode_size()
            self.__zoom_rate = zoom_rate
        elif fit_mode == self.FIT_MANUALLY:
            resized_size = [ int(val * self.__zoom_rate) for val in img.size  ]
        resized_img = img.resize(resized_size)
        return resized_img

    def toggle_draw_direction(self,is_right_to_left):
        self.__is_right_to_left = is_right_to_left

    def set_fit_mode(self,fit_mode):
        if (not fit_mode == self.FIT_AUTO and
            not fit_mode == self.FIT_MANUALLY and
            not fit_mode == self.FIT_TO_WIDTH and
            not fit_mode == self.FIT_TO_HEIGHT):
            raise 'not defined fit mode'
        self.__fit_mode = fit_mode
        self._cash_draw()

    def get_fit_mode(self):
        return self.__fit_mode

    def get_right_to_left_mode(self):
        return self.__is_right_to_left

    def set_right_to_left_mode(self,is_right_to_left):
        self.__is_right_to_left = is_right_to_left
        self.__update_box_alignment()
        self._cash_draw()

    def set_manual_zoom_rate(self,zoom_rate):
        self.__zoom_rate = zoom_rate

    def get_manual_zoom_rate(self):
        return self.__zoom_rate

    def get_dualscreen_mode(self):
        return self.__is_dual

    def set_dualscreen_mode(self,is_dual):
        self.__is_dual = is_dual
        self.__update_box_size()
        self.__update_box_alignment()
        self._cash_draw()

    def _cash_draw(self):
        if self.__cur_img or self.__fwd_img:
            self.redraw()

class ArchiveViewer:
    __namelist = None
    __cur_index = 0
    __archiver = None
    __terminate_signal =False
    def __init__(self,drawer,img_object_factory):
        self.drawer = drawer
        self.img_factory = img_object_factory
        self.cache = ImageCacheManager(self.__get_img)

    def terminate(self):
        self.cache.terminate()
        if self.__archiver:
            self.__archiver.close()

    def run(self,file_uri,start_page_number=0,end_page_number=None):
        self.__init_archiver(file_uri)
        self.__init_viewer(start_page_number,end_page_number)
        self.move_page_to(self.__cur_index)

    def __init_archiver(self,file_uri):
        self.__archiver = archive.get_archiver(file_uri)
        if not self.__archiver:
            raise 'not supported archiver'
            return

    def __init_viewer(self,start_page_number,end_page_number):
        self.__start_page = start_page_number
        self.__cur_index = start_page_number
        self.__namelist = self.__archiver.namelist()
        self.__total_page = end_page_number if end_page_number else len(self.__namelist)

    def get_page_number(self):
        return self.__cur_index

    def get_page_sum(self):
        return self.__total_page

    def move_first_page(self):
        self.move_page_to(self.__start_page)

    def move_last_page(self):
        self.move_page_to(self.__total_page - 1)

    def move_next_page(self,degree=1):
        next_page = self.__cur_index + self.__get_page_padd() * degree
        if self.__total_page> next_page:
            self.move_page_to( next_page )
        else:
            self.move_page_to( self.__total_page -1 )

    def move_prev_page(self,degree=1):
        prev_page = self.__cur_index - self.__get_page_padd() * degree
        if self.__start_page <= prev_page:
            self.move_page_to( prev_page )
        else:
            self.move_page_to( self.__start_page )

    def move_page_to(self,num):
        if num > self.__total_page -1 or num < self.__start_page:
            raise 'cannot move to invalid page number' 
        self.__cur_index = num

        cur = self.__get_img_with_cache(num)

        fwd = None

        if num + 1 < self.__total_page:
            fwd = self.__get_img_with_cache(num + 1)

        self.drawer.draw_image( cur,fwd )

    def __get_img_with_cache(self,num):
        self.__regist_prev_and_next_key(num)
        img = self.cache.get(num)
        if not img:
            while not self.__terminate_signal and not img:
                img = self.cache.get(num)
                time.sleep(0.001)
        return img

    def __get_page_padd(self):
        if not self.drawer.get_dualscreen_mode():
            padd = 1
        else:
            padd = 2
        return padd

    def __regist_prev_and_next_key(self,num):
        total_csize = config.AppConfig.PAGE_CACHE_SIZE
        prev_csize = total_csize / 2
        next_csize = total_csize - prev_csize
        processed_index = None

        def pickup_updated_keys(standerd):
            keyset = self.cache.keyset()
            nkey_end = standerd + next_csize
            if nkey_end > self.__total_page:nkey_end = self.__total_page

            pkey_start = standerd - prev_csize
            if pkey_start <0:pkey_start = 0

            pkey_range = range(pkey_start,standerd)
            nkey_range = range(standerd,nkey_end)
            
            #collect out-range keys to cache out them
            delkeys = keyset.difference(pkey_range + nkey_range)
            
            #collect new keys to cache in them
            p_newkeys = sorted(  set(pkey_range).difference(keyset) , reverse=True)
            n_newkeys = sorted(  set(nkey_range).difference(keyset) )
            return delkeys, p_newkeys, n_newkeys

        delkeys,prevs,nexts = pickup_updated_keys(num)
        self.cache.remove(delkeys)
        self.cache.regist(nexts)
        self.cache.regist(prevs)

    def __get_img(self,num):
        return self.img_factory.build( 
            self.__archiver.read( self.__namelist[num]) 
            )

class ImageCacheManager:
    __reqq = collections.deque()
    __cache = {}
    __active_thread = None
    __terminate_signal = False
    __processing_page = None

    def __init__(self,get_img_func):
        self.__get_img = get_img_func
        
    def regist(self,pages):
#        print 'reg:'+str(pages)
        if not self.__active_thread:
            self.__active_thread = threading.Thread(target=self.__worker)
            self.__active_thread.start()
        for page in pages:
            if (not page in self.__reqq and 
                not page in self.__cache and 
                page !=self.__processing_page):
                self.__reqq.appendleft(page)
#        print self.__reqq

    def remove(self,pages):
        for page in pages:
            if page in self.__cache:
                del self.__cache[page]
            elif page in self.__reqq:
                self.__reqq.remove(page)
#        print 'rm:'+str(pages)

    def terminate(self):
        if self.__active_thread:
            self.__terminate_signal = True
            self.__active_thread.join(0.5)
            
    def keyset(self):
        req_and_cache_set = set(self.__cache.keys()).union( self.__reqq)
        if self.__processing_page:
            req_and_cache_set = req_and_cache_set.union([self.__processing_page])
        return req_and_cache_set

    def get(self,page):
        return self.__cache.get(page,None)

    def __worker(self):
        try:
            while not self.__terminate_signal:
                
                if len(self.__reqq)==0:
                    continue

                self.__processing_page = self.__reqq.pop()
                page = self.__processing_page
                self.__cache[page] = self.__get_img(page)
        except:
            print sys.exc_info()
